import EventEmitter from 'eventemitter3';
export class Plan {
    promise;
    steps;
    eventEmitter;
    executing = false;
    executed = false;
    constructor(plan) {
        this.promise = plan.promise;
        this.steps = plan.steps ?? [];
        this.eventEmitter = plan.eventEmitter ?? new EventEmitter();
    }
    static make() {
        return new Plan({
            promise: async (initialState) => initialState,
        });
    }
    addStep(step) {
        const { newStep, handler } = this.parseInputStep(step);
        const promise = async (initialState, steps) => {
            let state;
            try {
                state = await this.promise(initialState, steps);
            }
            catch (error) {
                this.changeStepStatus(steps, newStep, 'canceled');
                throw error;
            }
            return this.processStep(steps, state, newStep, handler);
        };
        return new Plan({
            promise,
            steps: [...this.steps, newStep],
            eventEmitter: this.eventEmitter,
        });
    }
    prependStep(step) {
        const { newStep, handler } = this.parseInputStep(step);
        const promise = async (newInitialState, steps) => {
            let initialState;
            try {
                initialState = await this.processStep(steps, newInitialState, newStep, handler);
            }
            catch (error) {
                this.steps.forEach((step) => this.changeStepStatus(steps, step, 'canceled'));
                throw error;
            }
            return this.promise(initialState, steps);
        };
        return new Plan({
            promise,
            steps: [newStep, ...this.steps],
            eventEmitter: this.eventEmitter,
        });
    }
    parseInputStep(step) {
        const { handler } = step;
        const newStep = {
            name: step.name,
            status: 'pending',
            hidden: step.hidden ?? false,
            optional: step.optional ?? false,
            onError: step.onError,
        };
        return { newStep, handler };
    }
    async processStep(steps, from, step, handler) {
        this.changeStepStatus(steps, step, 'running');
        try {
            const to = await handler(from);
            this.changeStepStatus(steps, step, 'successful');
            return to;
        }
        catch (error) {
            this.changeStepStatus(steps, step, 'failed');
            step.onError?.(error);
            if (step.optional) {
                // If a step is optional, it's destination state should match
                // the source state. Otherwise, steps cannot be composed.
                return from;
            }
            else {
                throw error;
            }
        }
    }
    onChange(listener) {
        this.eventEmitter.addListener('change', listener);
        return this;
    }
    notifyChange(step, steps) {
        this.eventEmitter.emit('change', step, steps);
    }
    changeStepStatus(steps, step, newStatus) {
        step.status = newStatus;
        this.notifyChange(step, steps);
    }
    getSteps() {
        return this.steps;
    }
    getVisibleSteps() {
        return this.steps.filter((step) => !step.hidden);
    }
    async execute(initialState) {
        try {
            this.executing = true;
            this.executed = false;
            const state = initialState ?? undefined;
            return await this.promise(state, this.steps);
        }
        finally {
            this.executing = false;
            this.executed = true;
        }
    }
}
//# sourceMappingURL=Plan.js.map